from pathlib import Path

import bpy
import mrcfile
import numpy as np
import starfile
from PIL import Image

from ... import blender as bl
from databpy import AttributeTypes
import databpy
from .ensemble import Ensemble


class StarFile(Ensemble):
    def __init__(self, file_path):
        super().__init__(file_path)
        self.type = "starfile"

    @classmethod
    def from_starfile(cls, file_path):
        self = cls(file_path)
        self.data = self._read()
        self.star_type = None
        self.positions = None
        self.current_image = -1
        self._create_mn_columns()
        self.n_images = self._n_images()
        return self

    @classmethod
    def from_blender_object(cls, blender_object):
        self = cls(blender_object["starfile_path"])
        self.object = blender_object
        self.data = self._read()
        self.star_type = None
        self.positions = None
        self.current_image = -1
        self._create_mn_columns()
        self.n_images = self._n_images()
        bpy.app.handlers.depsgraph_update_post.append(self._update_micrograph_texture)
        return self

    @property
    def star_node(self):
        return bl.nodes.get_star_node(self.object)

    @property
    def micrograph_material(self):
        return bl.nodes.MN_micrograph_material()

    def _read(self):
        star = starfile.read(self.file_path)
        return star

    def _n_images(self):
        if isinstance(self.data, dict):
            return len(self.data)
        return 1

    def _create_mn_columns(self):
        # only RELION 3.1 and cisTEM STAR files are currently supported, fail gracefully
        if (
            isinstance(self.data, dict)
            and "particles" in self.data
            and "optics" in self.data
        ):
            self.star_type = "relion"
        elif "cisTEMAnglePsi" in self.data:
            self.star_type = "cistem"
        else:
            raise ValueError(
                "File is not a valid RELION>=3.1 or cisTEM STAR file, other formats are not currently supported."
            )

        # Get absolute position and orientations
        if self.star_type == "relion":
            df = self.data["particles"].merge(self.data["optics"], on="rlnOpticsGroup")

            # get necessary info from dataframes
            # Standard cryoEM starfile don't have rlnCoordinateZ. If this column is not present
            # Set it to "0"
            if "rlnCoordinateZ" not in df:
                df["rlnCoordinateZ"] = 0

            self.positions = df[
                ["rlnCoordinateX", "rlnCoordinateY", "rlnCoordinateZ"]
            ].to_numpy()
            pixel_size = df["rlnImagePixelSize"].to_numpy().reshape((-1, 1))
            self.positions = self.positions * pixel_size
            shift_column_names = [
                "rlnOriginXAngst",
                "rlnOriginYAngst",
                "rlnOriginZAngst",
            ]
            if all([col in df.columns for col in shift_column_names]):
                shifts_ang = df[shift_column_names].to_numpy()
                self.positions -= shifts_ang
            df["MNAnglePhi"] = df["rlnAngleRot"]
            df["MNAngleTheta"] = df["rlnAngleTilt"]
            df["MNAnglePsi"] = df["rlnAnglePsi"]
            df["MNPixelSize"] = df["rlnImagePixelSize"]
            try:
                df["MNImageId"] = (
                    df["rlnMicrographName"].astype("category").cat.codes.to_numpy()
                )
            except KeyError:
                try:
                    df["MNImageId"] = (
                        df["rlnTomoName"].astype("category").cat.codes.to_numpy()
                    )
                except KeyError:
                    df["MNImageId"] = 0.0

            self.data = df

        elif self.star_type == "cistem":
            df = self.data
            df["cisTEMZFromDefocus"] = (df["cisTEMDefocus1"] + df["cisTEMDefocus2"]) / 2
            df["cisTEMZFromDefocus"] = (
                df["cisTEMZFromDefocus"] - df["cisTEMZFromDefocus"].median()
            )
            self.positions = df[
                [
                    "cisTEMOriginalXPosition",
                    "cisTEMOriginalYPosition",
                    "cisTEMZFromDefocus",
                ]
            ].to_numpy()
            df["MNAnglePhi"] = df["cisTEMAnglePhi"]
            df["MNAngleTheta"] = df["cisTEMAngleTheta"]
            df["MNAnglePsi"] = df["cisTEMAnglePsi"]
            df["MNPixelSize"] = df["cisTEMPixelSize"]
            df["MNImageId"] = (
                df["cisTEMOriginalImageFilename"]
                .astype("category")
                .cat.codes.to_numpy()
            )

    def _convert_mrc_to_tiff(self):
        if self.star_type == "relion":
            micrograph_path = self.object["rlnMicrographName_categories"][
                self.star_node.inputs["Image"].default_value - 1
            ]
        elif self.star_type == "cistem":
            micrograph_path = self.object["cisTEMOriginalImageFilename_categories"][
                self.star_node.inputs["Image"].default_value - 1
            ].strip("'")
        else:
            return False

        # This could be more elegant
        if not Path(micrograph_path).exists():
            pot_micrograph_path = Path(self.file_path).parent / micrograph_path
            if not pot_micrograph_path.exists():
                if self.star_type == "relion":
                    pot_micrograph_path = (
                        Path(self.file_path).parent.parent.parent / micrograph_path
                    )
                    if not pot_micrograph_path.exists():
                        raise FileNotFoundError(
                            f"Micrograph file {micrograph_path} not found"
                        )
                else:
                    raise FileNotFoundError(
                        f"Micrograph file {micrograph_path} not found"
                    )
            micrograph_path = pot_micrograph_path

        tiff_path = Path(micrograph_path).with_suffix(".tiff")
        if not tiff_path.exists():
            with mrcfile.open(micrograph_path) as mrc:
                micrograph_data = mrc.data.copy()

            # For 3D data sum over the z axis. Probalby would be nicer to load the data as a volume
            if micrograph_data.ndim == 3:
                micrograph_data = np.sum(micrograph_data, axis=0)
            # Normalize the data to 0-1
            micrograph_data = (micrograph_data - micrograph_data.min()) / (
                micrograph_data.max() - micrograph_data.min()
            )

            if micrograph_data.dtype != np.float32:
                micrograph_data = micrograph_data.astype(np.float32)

            # Need to invert in Y to generate the correct tiff
            Image.fromarray(micrograph_data[::-1, :]).save(tiff_path)
        return tiff_path

    def _update_micrograph_texture(self, *_):
        try:
            show_micrograph = self.star_node.inputs["Show Micrograph"]
            _ = self.object["mn"]
        except ReferenceError:
            bpy.app.handlers.depsgraph_update_post.remove(
                self._update_micrograph_texture
            )
            return
        if self.star_node.inputs["Image"].default_value == self.current_image:
            return
        else:
            self.current_image = self.star_node.inputs["Image"].default_value
        if not show_micrograph:
            return
        tiff_path = self._convert_mrc_to_tiff()
        if tiff_path:
            try:
                image_obj = bpy.data.images[tiff_path.name]
            except KeyError:
                image_obj = bpy.data.images.load(str(tiff_path))
            image_obj.colorspace_settings.name = "Non-Color"
            self.micrograph_material.node_tree.nodes["Image Texture"].image = image_obj
            self.star_node.inputs["Micrograph"].default_value = image_obj

    def create_object(self, name="StarFileObject", node_setup=True, world_scale=0.01):
        self.object = databpy.create_object(
            self.positions * world_scale, collection=bl.coll.mn(), name=name
        )

        self.object.mn["entity_type"] = "star"

        # create attribute for every column in the STAR file
        for col in self.data.columns:
            col_type = self.data[col].dtype
            # If col_type is numeric directly add
            if np.issubdtype(col_type, np.number):
                self.store_named_attribute(
                    name=col,
                    data=self.data[col].to_numpy().reshape(-1),
                    atype=AttributeTypes.FLOAT,
                )

            # If col_type is object, convert to category and add integer values
            elif isinstance(col_type, object):
                codes = (
                    self.data[col].astype("category").cat.codes.to_numpy().reshape(-1)
                )
                self.store_named_attribute(
                    data=codes, name=col, atype=AttributeTypes.INT
                )
                # Add the category names as a property to the blender object
                self.object[f"{col}_categories"] = list(
                    self.data[col].astype("category").cat.categories
                )

        if node_setup:
            bl.nodes.create_starting_nodes_starfile(self.object, n_images=self.n_images)

        self.object["starfile_path"] = str(self.file_path)
        bpy.app.handlers.depsgraph_update_post.append(self._update_micrograph_texture)
        return self.object



import io
import time
import warnings
from abc import ABCMeta
from pathlib import Path
from typing import Optional, Tuple, Union

import biotite.structure as struc
import bpy
import numpy as np
import numpy.typing as npt
from biotite import InvalidFileError

from ... import blender as bl
from ... import color, data, utils
from databpy import Domains, AttributeTypes
import databpy
from ..entity import MolecularEntity, EntityType


class Molecule(MolecularEntity, metaclass=ABCMeta):
    """
    Abstract base class for representing a molecule.

    It associates the atomic data (the array) with the created 3D model inside of Blender
    (the object). If multiple conformations are imported, then a `frames` collection
    is also instantiated.

    The `named_attribute()` and `store_named_attribute()` methods access and set attributes on
    `object` that is in the Blender scene.

    Attributes
    ----------
    file_path : str
        The file path to the file which stores the atomic coordinates.
    file : Any
        The opened file.
    object : bpy.types.Object
        The Blender object representing the molecule.
    frames : bpy.types.Collection
        The Blender collection which holds the objects making up the frames to animate.
    array: np.ndarray:
        The numpy array which stores the atomic coordinates and associated attributes.
    entity_ids : np.ndarray
        The entity IDs of the molecule.
    chain_ids : np.ndarray
        The chain IDs of the molecule.

    Methods
    -------
    store_named_attribute(data, name='NewAttribute', type=None, domain='POINT', overwrite=True)
        Set an attribute on the object for the molecule.
    named_attribute(name='position')
        Get the value of an attribute on the object for the molecule.
    create_object(name='NewMolecule', style='spheres', selection=None, build_assembly=False, centre='', del_solvent=True, collection=None, verbose=False)
        Create a 3D model for the molecule, based on the values from self.array.
    assemblies(as_array=False)
        Get the biological assemblies of the molecule.
    """

    def __init__(self, file_path: Union[str, Path, io.BytesIO]):
        """
        Initialize the Molecule object.

        Parameters
        ----------
        file_path : Union[str, Path, io.BytesIO]
            The file path to the file which stores the atomic coordinates.
        """
        super().__init__()
        self._parse_filepath(file_path=file_path)
        self.file: str
        self.array: np.ndarray
        self._frames_collection: str | None
        self._entity_type = EntityType.MOLECULE

    @property
    def frames(self) -> bpy.types.Collection:
        """
        Get the collection of frames for the molecule.

        Returns
        -------
        bpy.types.Collection
            The collection of frames for the molecule.
        """
        if self._frames_collection is None:
            raise ValueError("No frames collection has been set for this molecule.")

        return bpy.data.collections[self._frames_collection]

    @frames.setter
    def frames(self, value: bpy.types.Collection):
        """
        Set the collection of frames for the molecule.

        Parameters
        ----------
        value : bpy.types.Collection
            The collection of frames for the molecule.
        """
        if value is None:
            self._frames_collection = None
            return
        if not isinstance(value, bpy.types.Collection):
            raise TypeError("The frames must be a bpy.types.Collection.")

        self._frames_collection = value.name

    @classmethod
    def _read(self, file_path: Union[Path, io.BytesIO]):
        """
        Initially open the file, ready to extract the required data.

        Parameters
        ----------
        file_path : Union[Path, io.BytesIO]
            The file path to the file which stores the atomic coordinates.
        """
        pass

    def _parse_filepath(self, file_path: Union[Path, str, io.BytesIO]) -> None:
        """
        If this is an actual file resolve the path - if a bytes IO resolve this as well.

        Parameters
        ----------
        file_path : Union[Path, str, io.BytesIO]
            The file path to the file which stores the atomic coordinates.
        """
        if isinstance(file_path, io.BytesIO):
            self.file = self._read(file_path=file_path)
        elif isinstance(file_path, io.StringIO):
            self.file = self._read(file_path=file_path)
        else:
            self.file_path = bl.path_resolve(file_path)
            self.file = self._read(self.file_path)

    def __len__(self) -> int:
        """
        Get the number of atoms in the molecule.

        Returns
        -------
        int
            The number of atoms in the molecule.
        """
        if hasattr(self, "object"):
            if self.object:
                return len(self.object.data.vertices)
        if self.array:
            return len(self.array)
        else:
            return 0

    @property
    def n_models(self):
        """
        Get the number of models in the molecule.

        Returns
        -------
        int
            The number of models in the molecule.
        """
        if isinstance(self.array, struc.AtomArray):
            return 1
        else:
            return self.array.shape[0]

    @property
    def tree(self) -> bpy.types.GeometryNodeTree:
        return self.object.modifiers["MolecularNodes"].node_group

    @property
    def chain_ids(self) -> Optional[list]:
        """
        Get the unique chain IDs of the molecule.

        Returns
        -------
        Optional[list]
            The unique chain IDs of the molecule, or None if not available.
        """
        if self.array:
            if hasattr(self.array, "chain_id"):
                return np.unique(self.array.chain_id).tolist()

        return None

    def create_object(
        self,
        name: str = "NewMolecule",
        style: str = "spheres",
        selection: np.ndarray = None,
        build_assembly=False,
        centre: str = "",
        del_solvent: bool = True,
        del_hydrogen: bool = False,
        collection=None,
        verbose: bool = False,
        color: Optional[str] = "common",
    ) -> bpy.types.Object:
        """
        Create a 3D model of the molecule inside of Blender.

        Creates a 3D model with one vertex per atom, and one edge per bond. Each vertex
        is given attributes which correspond to the atomic data such as `atomic_number` for
        the element and `res_name` for the residue name that the atom is associated with.

        If multiple conformations of the structure are detected, the collection attribute
        is also created which will store an object for each conformation, so that the
        object can interpolate between those conformations.

        Parameters
        ----------
        name : str, optional
            The name of the model. Default is 'NewMolecule'.
        style : str, optional
            The style of the model. Default is 'spheres'.
        selection : np.ndarray, optional
            The selection of atoms to include in the model. Default is None.
        build_assembly : bool, optional
            Whether to build the biological assembly. Default is False.
        centre : str, optional
            Denote method used to determine center of structure. Default is '',
            resulting in no translational motion being removed. Accepted values
            are `centroid` or `mass`. Any other value will result in default
            behavior.
        del_solvent : bool, optional
            Whether to delete solvent molecules. Default is True.
        del_hydrogen: bool, optional
            Whether to delete hydrogen atoms. Default is False.
        collection : str, optional
            The collection to add the model to. Default is None.
        verbose : bool, optional
            Whether to print verbose output. Default is False.
        color : Optional[str], optional
            The color scheme to use for the model. Default is 'common'.

        Returns
        -------
        bpy.types.Object
            The created 3D model, as an object in the 3D scene.
        """
        is_stack = isinstance(self.array, struc.AtomArrayStack)

        if selection:
            array = self.array[selection]
        else:
            array = self.array

        # remove the solvent from the structure if requested
        if del_solvent:
            mask = np.invert(struc.filter_solvent(array))
            if is_stack:
                array = array[:, mask]
            else:
                array = array[mask]

        if del_hydrogen:
            mask = array.element != "H"
            if is_stack:
                array = array[:, mask]
            else:
                array = array[mask]

        obj, frames = _create_object(
            array=array,
            name=name,
            centre=centre,
            style=style,
            collection=collection,
            verbose=verbose,
        )

        if style:
            bl.nodes.create_starting_node_tree(
                object=obj, coll_frames=frames, style=style, color=color
            )

        try:
            obj["entity_ids"] = self.entity_ids
        except AttributeError:
            obj["entity_ids"] = None

        try:
            obj["biological_assemblies"] = self.assemblies()
        except InvalidFileError:
            obj["biological_assemblies"] = None

        if build_assembly and style:
            bl.nodes.assembly_insert(obj)

        # attach the model bpy.Object to the molecule object
        self.object = obj
        # same with the collection of bpy Objects for frames
        self.frames = frames

        return obj

    def assemblies(self, as_array=False):
        """
        Get the biological assemblies of the molecule.

        Parameters
        ----------
        as_array : bool, optional
            Whether to return the assemblies as an array of quaternions. Default is False.

        Returns
        -------
        dict or None
            The biological assemblies of the molecule, as a dictionary of
            transformation matrices, or None if no assemblies are available.
        """
        try:
            assemblies_info = self._assemblies()
        except InvalidFileError:
            return None

        if isinstance(assemblies_info, dict) and as_array:
            return utils.array_quaternions_from_dict(assemblies_info)

        return assemblies_info

    def __repr__(self) -> str:
        """
        Get the string representation of the Molecule object.

        Returns
        -------
        str
            The string representation of the Molecule object.
        """
        return f"<Molecule object: {self.name}>"


def _create_object(
    array,
    name=None,
    centre="",
    style="spherers",
    collection=None,
    world_scale=0.01,
    color_plddt: bool = False,
    verbose=False,
) -> Tuple[bpy.types.Object, bpy.types.Collection]:
    import biotite.structure as struc

    frames = None
    is_stack = isinstance(array, struc.AtomArrayStack)

    try:
        mass = np.array(
            [
                data.elements.get(x, {}).get("standard_mass", 0.0)
                for x in np.char.title(array.element)
            ]
        )
        array.set_annotation("mass", mass)
    except AttributeError as e:
        print(e)

    def centre_array(atom_array, centre):
        if centre == "centroid":
            atom_array.coord -= databpy.centre(atom_array.coord)
        elif centre == "mass":
            atom_array.coord -= databpy.centre(atom_array.coord, weight=atom_array.mass)

    if centre in ["mass", "centroid"]:
        if is_stack:
            for atom_array in array:
                centre_array(atom_array, centre)
        else:
            centre_array(atom_array, centre)

    if is_stack:
        if array.stack_depth() > 1:
            frames = array
        array = array[0]

    if not collection:
        collection = bl.coll.mn()

    bonds_array = []
    bond_idx = []

    if array.bonds:
        bonds_array = array.bonds.as_array()
        bond_idx = bonds_array[:, [0, 1]]
        # the .copy(order = 'C') is to fix a weird ordering issue with the resulting array
        bond_types = bonds_array[:, 2].copy(order="C")

    # creating the blender object and meshes and everything
    bob = databpy.create_bob(
        name=name,
        collection=collection,
        vertices=array.coord * world_scale,
        edges=bond_idx,
    )

    # Add information about the bond types to the model on the edge domain
    # Bond types: 'ANY' = 0, 'SINGLE' = 1, 'DOUBLE' = 2, 'TRIPLE' = 3, 'QUADRUPLE' = 4
    # 'AROMATIC_SINGLE' = 5, 'AROMATIC_DOUBLE' = 6, 'AROMATIC_TRIPLE' = 7
    # https://www.biotite-python.org/apidoc/biotite.structure.BondType.html#biotite.structure.BondType
    if array.bonds:
        bob.store_named_attribute(
            data=bond_types,
            name="bond_type",
            atype=AttributeTypes.INT,
            domain=Domains.EDGE,
        )

    # The attributes for the model are initially defined as single-use functions. This allows
    # for a loop that attempts to add each attibute by calling the function. Only during this
    # loop will the call fail if the attribute isn't accessible, and the warning is reported
    # there rather than setting up a try: except: for each individual attribute which makes
    # some really messy code.

    # I still don't like this as an implementation, and welcome any cleaner approaches that
    # anybody might have.

    def att_atomic_number():
        atomic_number = np.array(
            [
                data.elements.get(x, {"atomic_number": -1}).get("atomic_number")
                for x in np.char.title(array.element)
            ]
        )
        return atomic_number

    def att_atom_id():
        return array.atom_id

    def att_res_id():
        return array.res_id

    def att_res_name():
        other_res = []
        counter = 0
        id_counter = -1
        res_names = array.res_name
        res_ids = array.res_id
        res_nums = []

        for name in res_names:
            res_num = data.residues.get(name, {"res_name_num": -1}).get("res_name_num")

            if res_num == 9999:
                if (
                    res_names[counter - 1] != name
                    or res_ids[counter] != res_ids[counter - 1]
                ):
                    id_counter += 1

                unique_res_name = str(id_counter + 100) + "_" + str(name)
                other_res.append(unique_res_name)

                num = (
                    np.where(np.isin(np.unique(other_res), unique_res_name))[0][0] + 100
                )
                res_nums.append(num)
            else:
                res_nums.append(res_num)
            counter += 1

        bob.object["ligands"] = np.unique(other_res)
        return np.array(res_nums)

    def att_chain_id():
        return np.unique(array.chain_id, return_inverse=True)[1]

    def att_entity_id():
        return array.entity_id

    def att_b_factor():
        return array.b_factor

    def att_occupancy():
        return array.occupancy

    def att_vdw_radii():
        vdw_radii = np.array(
            list(
                map(
                    # divide by 100 to convert from picometres to angstroms which is
                    # what all of coordinates are in
                    lambda x: data.elements.get(x, {}).get("vdw_radii", 100.0) / 100,
                    np.char.title(array.element),
                )
            )
        )
        return vdw_radii * world_scale

    def att_mass():
        return array.mass

    def att_atom_name():
        atom_name = np.array(
            list(map(lambda x: data.atom_names.get(x, -1), array.atom_name))
        )

        return atom_name

    def att_lipophobicity():
        lipo = np.array(
            list(
                map(
                    lambda x, y: data.lipophobicity.get(x, {"0": 0}).get(y, 0),
                    array.res_name,
                    array.atom_name,
                )
            )
        )

        return lipo

    def att_charge():
        charge = np.array(
            list(
                map(
                    lambda x, y: data.atom_charge.get(x, {"0": 0}).get(y, 0),
                    array.res_name,
                    array.atom_name,
                )
            )
        )
        return charge

    def att_color():
        if color_plddt:
            return color.plddt(array.b_factor)
        else:
            return color.color_chains(att_atomic_number(), att_chain_id())

    def att_is_alpha():
        return np.isin(array.atom_name, "CA")

    def att_is_solvent():
        return struc.filter_solvent(array)

    def att_is_backbone():
        """
        Get the atoms that appear in peptide backbone or nucleic acid phosphate backbones.
        Filter differs from the Biotite's `struc.filter_peptide_backbone()` in that this
        includes the peptide backbone oxygen atom, which biotite excludes. Additionally
        this selection also includes all of the atoms from the ribose in nucleic acids,
        and the other phosphate oxygens.
        """
        backbone_atom_names = [
            "N",
            "C",
            "CA",
            "H",  # backbone hydrogen off N
            "HA",  # backbone hydrogen off CA
            "O",  # peptide backbone atoms
            "P",
            "O5'",
            "C5'",
            "C4'",
            "C3'",
            "O3'",  # 'continuous' nucleic backbone atoms
            "O1P",
            "OP1",
            "O2P",
            "OP2",  # alternative names for phosphate O's
            "O4'",
            "C1'",
            "C2'",
            "O2'",  # remaining ribose atoms
        ]

        return np.logical_and(
            np.isin(array.atom_name, backbone_atom_names),
            np.logical_not(struc.filter_solvent(array)),
        )

    def att_is_nucleic() -> npt.NDArray[np.bool_]:
        return struc.filter_nucleotides(array)

    def att_is_peptide() -> npt.NDArray[np.bool_]:
        aa = struc.filter_amino_acids(array)
        con_aa = struc.filter_canonical_amino_acids(array)

        return aa | con_aa

    def att_is_side_chain() -> npt.NDArray[np.bool_]:
        not_backbone = np.logical_not(att_is_backbone())

        return np.logical_and(
            not_backbone, np.logical_or(att_is_nucleic(), att_is_peptide())
        )

    def att_is_hetero():
        return array.hetero

    def att_is_carb() -> npt.NDArray[np.bool_]:
        return struc.filter_carbohydrates(array)

    def att_sec_struct():
        return array.sec_struct

    # these are all of the attributes that will be added to the structure
    # TODO add capcity for selection of particular attributes to include / not include to potentially
    # boost performance, unsure if actually a good idea of not. Need to do some testing.
    attributes = (
        {"name": "res_id", "value": att_res_id, "type": "INT", "domain": "POINT"},
        {"name": "res_name", "value": att_res_name, "type": "INT", "domain": "POINT"},
        {
            "name": "atomic_number",
            "value": att_atomic_number,
            "type": "INT",
            "domain": "POINT",
        },
        {"name": "b_factor", "value": att_b_factor, "type": "FLOAT", "domain": "POINT"},
        {
            "name": "occupancy",
            "value": att_occupancy,
            "type": "FLOAT",
            "domain": "POINT",
        },
        {
            "name": "vdw_radii",
            "value": att_vdw_radii,
            "type": "FLOAT",
            "domain": "POINT",
        },
        {"name": "mass", "value": att_mass, "type": "FLOAT", "domain": "POINT"},
        {"name": "chain_id", "value": att_chain_id, "type": "INT", "domain": "POINT"},
        {"name": "entity_id", "value": att_entity_id, "type": "INT", "domain": "POINT"},
        {"name": "atom_id", "value": att_atom_id, "type": "INT", "domain": "POINT"},
        {"name": "atom_name", "value": att_atom_name, "type": "INT", "domain": "POINT"},
        {
            "name": "lipophobicity",
            "value": att_lipophobicity,
            "type": "FLOAT",
            "domain": "POINT",
        },
        {"name": "charge", "value": att_charge, "type": "FLOAT", "domain": "POINT"},
        {"name": "Color", "value": att_color, "type": "FLOAT_COLOR", "domain": "POINT"},
        {
            "name": "is_backbone",
            "value": att_is_backbone,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {
            "name": "is_side_chain",
            "value": att_is_side_chain,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {
            "name": "is_alpha_carbon",
            "value": att_is_alpha,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {
            "name": "is_solvent",
            "value": att_is_solvent,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {
            "name": "is_nucleic",
            "value": att_is_nucleic,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {
            "name": "is_peptide",
            "value": att_is_peptide,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {
            "name": "is_hetero",
            "value": att_is_hetero,
            "type": "BOOLEAN",
            "domain": "POINT",
        },
        {"name": "is_carb", "value": att_is_carb, "type": "BOOLEAN", "domain": "POINT"},
        {
            "name": "sec_struct",
            "value": att_sec_struct,
            "type": "INT",
            "domain": "POINT",
        },
    )

    # assign the attributes to the object
    for att in attributes:
        if verbose:
            start = time.process_time()
        try:
            bob.store_named_attribute(
                data=att["value"](),
                name=att["name"],
                atype=att["type"],
                domain=att["domain"],
            )
            if verbose:
                print(f'Added {att["name"]} after {time.process_time() - start} s')
        except Exception as e:
            if verbose:
                print(e)
                warnings.warn(f"Unable to add attribute: {att['name']}")
                print(
                    f'Failed adding {att["name"]} after {time.process_time() - start} s'
                )

    coll_frames = None
    if frames:
        coll_frames = bl.coll.frames(bob.name)
        for i, frame in enumerate(frames):
            frame = databpy.create_object(
                name=bob.name + "_frame_" + str(i),
                collection=coll_frames,
                vertices=frame.coord * world_scale,
            )

    # add custom properties to the actual blender object, such as number of chains, biological assemblies etc
    # currently biological assemblies can be problematic to holding off on doing that
    try:
        bob.object["chain_ids"] = list(np.unique(array.chain_id))
    except AttributeError:
        bob.object["chain_ids"] = None
        warnings.warn("No chain information detected.")

    return bob.object, coll_frames



import molecularnodes as mn
import molecularnodes.blender as bl
import pytest
import bpy

from .constants import data_dir, attributes
from .utils import NumpySnapshotExtension


formats = ["mol", "sdf"]


@pytest.mark.parametrize("format", formats)
def test_open(snapshot_custom, format):
    molecule = mn.entities.molecule.SDF(data_dir / f"caffeine.{format}")

    assert molecule.array
    assert molecule.file


@pytest.mark.parametrize("format", formats)
@pytest.mark.parametrize("style", ["ball_and_stick", "spheres", "surface"])
def test_load(snapshot_custom: NumpySnapshotExtension, format, style):
    mol = mn.entities.load_local(data_dir / f"caffeine.{format}", style=style)
    assert mol.object

    if style == "spheres":
        bl.nodes.get_style_node(mol.object).inputs[
            "Sphere As Mesh"
        ].default_value = True
    mn.blender.nodes.realize_instances(mol.object)

    for attribute in attributes:
        try:
            assert snapshot_custom == mol.named_attribute(attribute, evaluate=True)
        except AttributeError as e:
            assert e
